/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

tree_init ()  /* initializes all constant nodes */
{
	Tree_init_constants ();
	pass_init ();
}

Node *tree;

int tree_list_len (list)
Node *list;
{
	Node *ele, *temp;
	int len = 0;

	if (list == NN)
		return 0;

	for (LIST_LOOP (ele, list, temp))
		len++;

	return len;
}

Node *list_add (last, value)  /* adds AFTER the last */
Node *last, *value;
{
	Node *newlast;

	if (last != NN) {
		newlast = make (SUBSCRIPT_LIST);
		newlast->lst.next = last->lst.next;
		last->lst.next = newlast;
		newlast->lst.Comma = &Comma;
		newlast->lst.value = value;
	}
	else {
		newlast = make (SUBSCRIPT_LIST);
		newlast->lst.Comma = &Comma;
		newlast->lst.value = value;
	}
	
	return newlast;
}

Node *list_add_end (list, value, type)  /* adds like during parsing */
Node *list, *value;
int type;
{
	Node *newlist;

	newlist = make (type);
	newlist->lst.next = list;
	newlist->lst.value = value;
	newlist->lst.Comma = &Comma;

	return newlist;
}

Node *list_add_beg (list, value, type)  /* adds at the beginning */
Node *list, *value;
int type;
{
	Node *ele, *temp, *newlist;

	if (list == NN) {
		newlist = make (type);
		newlist->lst.value = value;

		return newlist;
	}

	for (LIST_LOOP (ele, list, temp))
		if (LIST_LAST (temp)) {
			newlist = make (type);
			temp->lst.next = newlist;
			newlist->lst.value = value;
			temp->lst.Comma = &Comma;

			return list;
		}
}

/*!!! remember to freed the old list if unneeded */

Node *list_invert (list)  /* creates a new, inverted list */
Node *list;
{
	Node *ele, *nlist, *chain, *temp;
	int type;

	if (list == (Node*) 0)
		return (Node*) 0;

	type = list->com.type;

	for (LIST_LOOP (ele, list, temp)) {
		if (LIST_FIRST (ele, list)) {
			chain = make (type);
			chain->lst.value = ele;
		}
		else {
			nlist = make (type);
			nlist->lst.value = ele;
			nlist->lst.next = chain;
			chain = nlist;
		}
	}

	return chain;
}

Node *make (type)  /* make node of type, all but type field set to binary zero */
int type;
{
	Node *node;
	int size = node_size (type);

	node = (Node*) U_malloc (size);

	node->com.type = type;

	return node;
}

tree_pp (node)  /* pretty-prints tree starting at node */
Node *node;
{
	int i;
	int type, child_type;
	Node *child;

	if (node == (Node*) 0)
		return;

	type = node->com.type;

	if (node_class (type) != NONTERM)
		goto walk;

	printf ("%lx: %s -->", node, node_Symbol (type));

	if (node->com.typ != (Typ*) 0) {
		printf ("{");
		typ_show (node->com.typ);
		printf ("}");
	}

	for (i = 0; i < node_nodes (type); i++)
		if ((child = node->com.node[i]) != (Node*) 0) {
			child_type = child->com.type;

			switch (node_class (child_type)) {

			case NONTERM:
				printf (" [%s %lx]", node_Symbol (child_type), child);
				break;

			case TERM:

				printf (" [");
				if (child->com.typ != (Typ*) 0) {
					printf ("{");
					typ_show (child->com.typ);
					printf ("} ");
				}
			
				if (child->com.str == (char*) 0) {
					printf ("%s]", node_Symbol (child_type));
				}
				else {
					if (strlen (child->com.str) == 1 && *child->com.str == '\n')
						printf ("%s \"\\n\"]", node_Symbol (child_type));
					else
						printf ("%s \"%s\"]", node_Symbol (child_type), child->com.str);
				}
				break;

			case LIT:
				printf (" ['%c']", child_type);
				break;

			default:
				sprintf (serr, "type %d class %d", child_type, node_class (child_type));
				syserr (serr);
			}
		}
		else
			printf (" []");

	printf (".\n");

walk:
	for (i = 0; i < node_nodes (node->com.type); i++)
		tree_pp (node->com.node[i]);
}

tree_free (node)  /* frees tree starting at node */
Node *node;
{
	int i;

	for (i = 0; i < node_nodes (node->com.type); i++)
		tree_free (node->com.node[i]);

	U_free (node);
}

Node *tree_dup (node)  /* creates duplicate of tree starting at node */
Node *node;
{
	int i, type;
	Node *new;

	if (node == (Node*) 0)
		return (Node*) 0;

	type = node->com.type;

	switch (node_class (type)) {

	case LIT:
	case TERM:  /*!!! actually here we may need to copy in a few cases (NAME, ...) */
		new = node;
		break;

	case NONTERM:
		new = make (type);
		new->com.type = node->com.type;
		new->com.str = node->com.str;  /*!!! this causes trouble with tree_free */
		new->com.typ = node->com.typ;  /*!!! this causes trouble with tree_free */

		for (i = 0; i < node_nodes (type); i++)
			new->com.node[i] = tree_dup (node->com.node[i]);
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	return new;
}

tree_eq (n1, n2)  /* equal if types equal and if TERM with str != (char*) 0 also eq */
Node *n1, *n2;
{
	int i, type;

	if (n1 == (Node*) 0 && n2 == (Node*) 0)
		return 1;

	if (n1 == (Node*) 0 || n2 == (Node*) 0)
		return 0;

	if ((type = n1->com.type) != n2->com.type)
		return 0;

	switch (node_class (type)) {

	case LIT:
		return 1;
		break;

	case TERM:
		if (n1->com.str == (char*) 0 && n2->com.str == (char*) 0)
			return 1;

		if (n1->com.str == (char*) 0 || n2->com.str == (char*) 0)
			return 0;

		if (!strcmp (n1->com.str, n2->com.str))
			return 1;
		else
			return 0;

		break;

	case NONTERM:
		for (i = 0; i < node_nodes (type); i++)
			if (!tree_eq (n1->com.node[i], n2->com.node[i]))
				return 0;
		return 1;
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}
}
/******************************** P A S S ****************************************/
#include "pass.h"

pass_init_set_to_null ()
{
	int i, j, k;

	for (i = 0; i < PASS_TOTAL; i++)
		for (j = 0; j < 2; j++)
			for (k = 0; k < LAST_TOKEN; k++)
				function [i][j][k] = (PtFrN) 0;
}

int tree_has (node, this_type)  /* finds out if tree has node of type 'type' */
Node *node;
int this_type;
{
	int type, i;

	if (node == (Node*) 0)
		return 0;

	type = node->com.type;

	if (type == this_type)
		return 1;

	switch (node_class (type)) {

	case LIT:
	case TERM:
		return 0;
		break;

	case NONTERM:
		for (i = 0; i < node_nodes (type); i++)
			if (tree_has (node->com.node[i], this_type))
				return 1;

		return 0;

#ifdef CHECK
		default:
			SYSERR;
			return 0;
#endif
		}
}
